(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontbind.ml,v 1.25 2002/05/27 15:08:33 malekith Exp $
 *
 * Module implementation: Gontbind
 * Description:
 * The purpose of this file to make other part of compiler 
 * symbol-table-free (however type variables are handled in
 * Gonttypes, this module doesn't touch it). It generally changes
 * symbol options from None to Some sym. We also set variable
 * stating how deep in closure stack should we go to finally
 * get to the variable. We also mark variables to be put into
 * closures here.
 *)

open Gonttree
open Gontutil

(*
 * Level stuff:
 *   fnc_level is level of function nesting for symbol.
 *     -2 for undefined symbols
 *     -1 for symbols from other modules
 *      0 for global symbols (in current module)
 *      1 for variables and parameters to 1st level functions
 *      2 ditto for 2nd level...
 *   scope_level is level of scope nesting of symbol, levels -2 up to
 *   0 are the same as for fnc_level, next define level of block nesting.
 *
 * Example            // (fnc_level, scope_level)
 * int foo;           // (0, 0)
 * int bar(           // (0, 0)
 *     int x) {       // (1, 1)
 *     int baz;       // (1, 1)
 *     int f() {      // (1, 1)
 *         int k;     // (2, 2)
 *     }
 *     { int z; }     // (1, 2)
 *  }
 *)

type ctx = {ctx_types: typedef list;
            ctx_syms: symbol list;
            ctx_fnc_level: int;
            ctx_scope_level: int;
            ctx_opened_modules: string list;
            ctx_module: string;
            ctx_iface: bool;
            ctx_loop_label: id;
            ctx_init_body: stmt list;
            ctx_fini_body: stmt list}

let make_symbol c t n =
  {sym_scope_level = c.ctx_scope_level;
   sym_fnc_level = c.ctx_fnc_level;
   sym_type = t;
   sym_name = n;
   sym_kind = VariableKind;
   sym_properties = [];
   sym_typedefs = [];
  }
;;

let used_modules : (string, unit) Hashtbl.t = Hashtbl.create 16

let init_bind modname iface =
  Hashtbl.clear used_modules;
  {ctx_syms = []; 
   ctx_types = [];
   ctx_fnc_level = 0; 
   ctx_scope_level = 0;
   ctx_opened_modules = [];
   ctx_module = modname;
   ctx_iface = iface;
   ctx_loop_label = "$";
   ctx_init_body = [];
   ctx_fini_body = []}

let used_modules_list _ =
  Hashtbl.fold (fun s () lst -> s :: lst) used_modules []

let use_module m =
  if Hashtbl.mem used_modules m then
    m
  else begin
    Hashtbl.add used_modules m ();
    m
  end
  
let find_mod_type ctx m n =
     ignore(ctx);
     Gontmodule.import_type_symbol (use_module m) n
;;

(* If m = "", look in current namespace, and all open modules. *)
let rec find_sym ctx kind m n =
  let find_local () =
    let qn = (ctx.ctx_module ^ "::" ^ n) in
    List.find (fun sym -> (sym.sym_name = n || sym.sym_name = qn) &&
                           sym.sym_kind = kind) ctx.ctx_syms in
  if m = ctx.ctx_module then
    if ctx.ctx_iface then
      find_local ()
    else
      try
        find_local ()
      with Not_found -> 
        Gontmodule.import_symbol kind (use_module m) n
  else
    if m = "" then
      try
        find_local ()
      with Not_found -> 
        let rec find_in_mods mods =
          match mods with
          | x::xs -> 
            (try 
              find_sym ctx kind x n
             with Not_found ->
              find_in_mods xs)
          | [] -> raise Not_found in
        find_in_mods ctx.ctx_opened_modules
     else
       (let sym = Gontmodule.import_symbol kind (use_module m) n in
        set_property sym IsNonLocal;
        sym)
  
let find_type ctx m n =
  match m with
      "" -> 
        (try
          List.find (fun {td_name = x} -> n = x) ctx.ctx_types
         with Not_found ->
           let m = 
             List.find (fun m -> 
               try 
                 ignore(find_mod_type ctx m n); 
                 true 
               with Not_found -> false) ctx.ctx_opened_modules in
           find_mod_type ctx m n)
    | _ -> find_mod_type ctx m n

let add_symbol c loc kind t n =
  let (fnc_lev, scope_lev) =
    try
      let sym = find_sym c kind "" n in
      (sym.sym_fnc_level, sym.sym_scope_level)
    with Not_found -> (-2, -2) in
  let _ =
    match scope_lev with
    | x when (x == c.ctx_scope_level) ->
       err (loc, xf "redefinition of `%s' in the same scope" n)
    | x when (x > 0) ->
       warn (loc, xf "`%s' shadows other local symbol" n)
    | 0 ->
       warn (loc, xf "`%s' shadows other, global symbol" n)
    | -1 when (sym_module (find_sym c kind "" n) <> c.ctx_module) -> 
       warn (loc, xf "`%s' shadows symbol from other module" n)
    | _ -> () in
  let sym = make_symbol c t n in
  sym.sym_kind <- kind;
  ({c with ctx_syms = sym :: c.ctx_syms}, sym)
;;

let prepare_ctx loc (c, parms) =
  let rec helper (c, parms) =
    match parms with
        ({parm_type=t; parm_name=n} as x)::xs -> 
          let (c, sym) = add_symbol c loc VariableKind t n in
          set_property sym IsInitialized;
          set_property sym IsFormalParm;
          x.parm_symbol <- Some sym;
          helper (c, xs)
      | [] -> c in
  helper ({c with ctx_fnc_level   = c.ctx_fnc_level + 1;
                  ctx_scope_level = c.ctx_scope_level + 1}, 
          parms)
;;

(* This does in place modification of s. Updated context is returned. *)
let rec bind_stmt (c, s) =
  let rec bind_patlist (c, patl) =
    match patl with
        x::xs ->
          let {cc_pattern = pat; cc_stmt = st} = x in
          let pat_scope = 
            {c with ctx_scope_level = c.ctx_scope_level + 1} in
          let c' = bind_pattern (pat_scope, pat) in
          let c'' = bind_stmt (c', st) in
          ignore(c'');
          bind_patlist (c, xs)
      | [] -> c in
  match s.raw_stmt with
      Skip -> c
    | Exp(e) -> bind_exp (c, e)
    | VarDef v ->
          (match v.var_symbol with
              None -> ()
            | Some _ -> assert false);
          let (c, t) = bind_type (c, v.var_type) in
          let (c, sym) = add_symbol c s.stmt_loc VariableKind t v.var_name in
          v.var_symbol <- Some sym;
          c
    | Raise e
    | Return e -> bind_exp (c, e)
    | ReturnVoid -> c
    | Block(ss) ->
        let rec bind_block (c, ss) =
          match ss with
              x::xs -> 
                let c' = bind_stmt (c, x) in
                bind_block (c', xs)
            | [] -> c in
        let new_scope = {c with ctx_scope_level = c.ctx_scope_level + 1} in
        let c' = bind_block(new_scope, ss) in
        ignore(c'); 
        c
    | Try t ->
      let _ = bind_stmt (c, t.try_body) in
      (* with { .. } is bind in context untouched by try_body (we do not know
       * from which place in body we jump in here, and which variables might
       * be defined). *)
      let _ = bind_patlist (c, t.try_with) in
      (* Similar can be told about finally. *)
      let _ = bind_stmt (c, t.try_finally) in
      c 
    | Pair(s1, s2) ->
        bind_stmt ((bind_stmt (c, s1)), s2)
    | Switch(expr, patlist) ->
        let exprc = bind_exp (c, expr) in
        let patc = bind_patlist (exprc, patlist) in
        ignore(patc);
        exprc                   
    | Let (patel, stmt) ->
        let rec expr_binder (c, pel) =
           match pel with
           | x::xs ->
              let (_, ex) = x in
              let cn = bind_exp (c, ex) in
              expr_binder (cn, xs)
           | [] -> c in
        let rec pate_binder (c, pel) =
           match pel with
           | x::xs ->
              let (pa, _) = x in
              let cn = bind_let_pattern (c, pa) in
              expr_binder (cn, xs)
           | [] -> c in
        let c1 = expr_binder (c, patel) in
        let newscope =
          {c1 with ctx_scope_level = c1.ctx_scope_level + 1} in
        let c2 = pate_binder (newscope, patel) in
        let c3 = bind_stmt (c2, stmt) in
        ignore(c3);
        c1
    | BreakIfNot e ->
        bind_exp (c, e)
    | Loop l ->
        let (c', sym) = add_symbol c s.stmt_loc LoopIdKind 
                                   AbstractType l.ll_label in
        l.ll_sym <- Some sym;
        let c' = {c' with ctx_loop_label = l.ll_label} in
        let c' = bind_stmt (c', l.ll_body) in
        let c' = bind_stmt (c', l.ll_inc) in
        (* Now we can assign some name to this symbol... *)
        sym.sym_name <- seq_name "anon_loop$";
        (* ...and forget it :] (not really, we simply wipe it out from
         * binding context) *)
        ignore(c');
        c
    | Break b ->
        let n = if b.br_continue then "continue" else "break" in
        if c.ctx_loop_label = "$" then
            err (s.stmt_loc, xf "%s outside loop" n)
        else
          let label = 
            if b.br_label = "" && c.ctx_loop_label <> "" then
              let _ = warn (s.stmt_loc, xf "%s without label, while current \
                                            loop is labeled `%s'" 
                                           n c.ctx_loop_label) in
              c.ctx_loop_label
            else
              b.br_label in
          let _ = 
            try
              let sym = find_sym c LoopIdKind c.ctx_module label in
              b.br_sym <- Some sym
            with Not_found ->
              err (s.stmt_loc, xf "no loop enclosing this %s is labeled `%s'" 
                                  n label) in
        c
    | If(e, s1, s2) ->
        bind_stmt (bind_stmt (bind_exp (c, e), s1), s2)
    | Open m ->
        Gontmodule.open_module m;
        {c with ctx_opened_modules = m :: c.ctx_opened_modules}
and bind_exp (c, e) =
  match e.raw_exp with
      SimpleBinOp(op, e1, e2) -> 
        let c' = bind_exp (c, e1) in
        bind_exp (c', e2)
    | SimpleUnOp(op, e1) ->
        bind_exp (c, e1)
    | Ref(r) ->
        (match r.ref_symbol with
            None ->
              let sym = 
                try
                  find_sym c VariableKind "" r.ref_name
                with Not_found ->
                  err (e.exp_loc, xf "unbound symbol `%s'" r.ref_name) in
              r.ref_symbol <- Some sym;
              let lev = c.ctx_fnc_level - sym.sym_fnc_level in
                assert (lev >= 0);
                r.ref_level <- if sym.sym_fnc_level == 0 then 0 else lev
          (* We have already been here! *)
          | Some _ -> assert false); 
          (* Context is not updated in any way. *)
          c
    | CrossModuleRef(r) ->
        (match r.xmr_symbol with
            None ->
              (try
                 let sym = find_sym c VariableKind 
                                      r.xmr_module r.xmr_name in
                 r.xmr_symbol <- Some sym
               with Not_found ->
                 err (e.exp_loc, xf "there is no symbol %s in module %s"
                                    r.xmr_name r.xmr_module))
          | Some _ -> assert false);
          c
    | Call(e1, es) -> 
        let c' = bind_exp (c, e1) in
        let rec bind_parms (c, pp) =
            match pp with
              p::ps -> 
                let c' = bind_exp (c, p) in
                bind_parms (c', ps)
           | [] -> c in
        bind_parms (c', es)
    | IntConst _
    | BoolConst _
    | FloatConst _
    | StringConst _ 
    | NullConst
    | VoidConst -> c
    | UnionExp (un) -> 
        let _ =
            match un.ur_sym with
            | None ->
                (* find out if there exists union containing id field *)
                let sym = 
                  try
                    find_sym c VariableKind un.ur_module un.ur_name
                  with Not_found -> 
                   err (e.exp_loc, xf "there is no union field %s::%s"
                                      un.ur_module un.ur_name) in
                 un.ur_sym <- Some sym
            | Some _ -> assert false in
        bind_exp(c, un.ur_exp)
    | TupleExp (tup) ->
        let pair (ct, ex) = (bind_exp(ct,ex), ex) in
        let (c', _) = list_apply pair (c, tup) in
        c'
    | Lambda {lambda_parms=p; lambda_ret_type=r; lambda_body=s} ->
        let new_ctx = prepare_ctx e.exp_loc (c, p) in
        let (c', t) = bind_type (new_ctx, make_fn_type (r, p)) in
        let c'' = bind_stmt (c', s) in
        ignore(c'');
       (* We return the same context. Local function cannot affect
        * parent environment. *)
       c
    | StructExp sis ->
        let bind_si c si =
          match si with
          | {si_sym = None} as si ->
              let sym = 
                try
                  find_sym c StructFieldKind si.si_module si.si_name
                with Not_found ->
                  err (si.si_loc, xf "unbound struct field `%s::%s'" 
                                     si.si_module si.si_name) in
              si.si_sym <- Some sym;
              bind_exp (c, si.si_expr)
          | {si_sym = Some _} -> assert false in
          let c' = List.fold_left bind_si c sis in
          ignore(c');   (* ({ ... }) creates scope. *)
          c
    (* We cannot tell here, wich symbol to put in de_sym, until we'll find
     * out type of expression, so we just bind expression inside. *)
    | DotExp de ->
        bind_exp (c, de.de_expr)
and add_fields (c, td) =
  match td.td_type with
    Union uni ->       (* executed during union type declaration *)
      let rec bind_union_fields (no, c, fil) =  
        match fil with
          x::xs ->
            let {fr_type = t; fr_name = id; fr_module = m} = x in
            if m <> "" && m <> c.ctx_module then
              err(td.td_loc, xf "module has to be `%s' or empty, not `%s'" 
                                c.ctx_module m);
            let (c, sym) = add_symbol c td.td_loc VariableKind t id in
            sym.sym_name <- (c.ctx_module ^ "::" ^ id);
            sym.sym_properties <- (UnionFieldId no) :: sym.sym_properties;
            sym.sym_typedefs <- [td];
            set_property sym IsUnionField;
            (if c.ctx_iface then
              x.fr_module <- c.ctx_module);
            (*x.fr_sym <- Some sym;*)
            bind_union_fields(no + 1, c, xs)
        | [] -> c in
      bind_union_fields (1, c, uni)
    | Struct l
    | OptStruct l ->
      let bind_struct_field c f =
         let {fr_type = t; fr_name = id; fr_module = m} = f in
         if m <> "" && m <> c.ctx_module then
           err(td.td_loc, xf "module has to be `%s' or empty, not `%s'" 
                             c.ctx_module m);
         let sym =
           try
             find_sym c StructFieldKind c.ctx_module id
           with Not_found ->
             (let s = 
               make_symbol c AbstractType (c.ctx_module ^ "::" ^ id) in
             s.sym_kind <- StructFieldKind;
             s) in
         if List.exists (fun t -> t == td) sym.sym_typedefs then
           err(f.fr_loc, xf "field %s already defined in %s"
                            sym.sym_name td.td_name);
         sym.sym_typedefs <- td :: sym.sym_typedefs;
         if (List.memq sym c.ctx_syms) then
           c
         else
           {c with ctx_syms = sym :: c.ctx_syms} in
       List.fold_left bind_struct_field c l
    | _ -> c
      
and bind_type (c, t) =
  let add_sym kind fr =
    match fr.fr_sym with
    | Some _ -> fr
    | None ->
      let sym =
        try
          find_sym c kind fr.fr_module fr.fr_name
        with Not_found ->
          err(fr.fr_loc, xf "unbound record field %s::%s" 
                            fr.fr_module fr.fr_name) in
      fr.fr_sym <- Some sym;
      fr in
  let f t =
    match t with
        NamedType ({nt_tydef = None} as r) ->
          (try
             r.nt_tydef <- Some (find_type c r.nt_module r.nt_name);
             NamedType r
           with Not_found ->
             if r.nt_module = "" then
               err (r.nt_loc, xf "use of unknown type %s" r.nt_name)
             else
               err (r.nt_loc, xf "module %s does not define type %s"
                                 r.nt_module r.nt_name))
        (* We used to have assert false here (as we already bound this
         * type), however lambda expressions generated by parsers
         * are somewhat tricky -- they use the same type for type of
         * functional variable being defined and for parameters of
         * functional value, so bind_type() finishes up called two times
         * on the same type. This should be probably fixed in parser,
         * not here, but I'm not sure. --mal *)
      | NamedType _ -> t
      | Union l -> Union (List.map (add_sym VariableKind) l)
      | Struct l -> Struct (List.map (add_sym StructFieldKind) l) 
      | OptStruct l -> OptStruct (List.map (add_sym StructFieldKind) l)
      | Exn | Int | Void | Bool | Float | String | TyVar _ | AbstractType
      | Tuple _ | Func _ -> t in
  (c, type_walk f t)
and bind_pattern (c, p) =
  match p.raw_pattern with
    VoidPattern -> 
      c
  | VarPattern vap ->
      (match vap.vp_symbol with
        None ->
          let sym = make_symbol c AbstractType vap.vp_name in
          set_property sym IsInitialized;
          vap.vp_symbol <- Some sym;
          {c with ctx_syms = sym :: c.ctx_syms}  
      (* We have already been here! *)
      | Some _ -> assert false
      )
  | ExprPattern e ->
      (* 
       * Non-functional voodoo magic:
       *
       * `case foo:' might be 3 different things:
       *   1) reference to constant variable
       *   2) union (or expression) constructor
       *   3) free pattern variable
       * We do not know it while parsing input file, we cannot tell until
       * here. So we change the patter expression to reflect reality. 
       *)
      let c =
        match e.raw_exp with
        | Ref r ->
          begin
            try
              (* If we can find variable symbol for it, ... *)
              let _ = find_sym c VariableKind "" r.ref_name in
              (* ...simply bind it. *)
              bind_exp(c, e)
            with Not_found ->
              (* Otherwise it's free pattern variable. *)
              let _ = p.raw_pattern <- VarPattern {vp_name = r.ref_name;
                                                   vp_symbol = None} in
              bind_pattern(c, p)
          end
        | _ -> bind_exp (c, e) in
      begin
        (* Ok, we bound it somewhat. *)
        match e.raw_exp with
        (* Check if it's union/exn ctor. *)
        | Ref {ref_symbol = Some sym}
        | CrossModuleRef {xmr_symbol = Some sym}
            when (has_property sym IsUnionField ||
                  has_property sym IsException) ->
            (* Yap, change to UnionPattern. *)
            let nil = {raw_pattern = TuplePattern {tpr_patterns = []};
                       pat_loc = e.exp_loc} in
            let _ = p.raw_pattern <- 
                           UnionPattern {upr_module = "";
                                         upr_name = sym.sym_name;
                                         upr_sym = Some sym;
                                         upr_pattern = nil} in
            c
        (* Nope, return c. *)
        | _ -> c
      end
  | TuplePattern tup ->
      let pair (ct, pa) = (bind_pattern (ct, pa), pa) in
      let (c', _) = list_apply pair (c, tup.tpr_patterns) in
      c'
  | UnionPattern un ->
      let sym = 
        try
          find_sym c VariableKind un.upr_module un.upr_name 
        with Not_found -> 
          (*List.iter(fun s -> print_endline s#get_name) c.ctx_syms; *)
          err(p.pat_loc, xf "union field `%s::%s' not defined" 
                            un.upr_module un.upr_name) in
      un.upr_sym <- Some sym;
      bind_pattern (c, un.upr_pattern)

and bind_let_pattern (c, p) =
  match p.raw_pattern with
    VoidPattern ->
      c
  | VarPattern vap ->
      (match vap.vp_symbol with
        None ->
          let (c, sym) = add_symbol c p.pat_loc VariableKind 
                                                AbstractType vap.vp_name in
          vap.vp_symbol <- Some sym;
          set_property sym IsInitialized;
          c  
      (* We have already been here! *)
      | Some _ -> assert false
      )
  | ExprPattern e ->
      (match e.raw_exp with
      | Ref r ->
          p.raw_pattern <- VarPattern {vp_name = r.ref_name;
                                       vp_symbol = None};
          p.pat_loc <- e.exp_loc;
          bind_let_pattern(c, p)
       | _ -> 
         err (e.exp_loc, "no expressions allowed in let pattern")
      )
  | TuplePattern tup ->
      let pair (ct, pa) = (bind_let_pattern (ct, pa), pa) in
      let (c', _) = list_apply pair (c, tup.tpr_patterns) in
      c'
  | UnionPattern un ->
      err (p.pat_loc, "no unions allowed in let pattern")
      (* although there could be union with one field, but 
         I think we don't have to consider it *)  
;;

(* Modify stmts in place, adding symbol entries to Ref terms. This mustn't be
 * called more then once in a top level function. *)
let bind c decl =
  match decl with
      FuncDecl f ->
        let fntype = make_fn_type (f.fn_ret_type, f.fn_parms) in
        let (c, fntype) = bind_type (c, fntype) in
        let (c, sym) = add_symbol c f.fn_loc VariableKind fntype f.fn_name in
        sym.sym_name <- c.ctx_module ^ "::" ^ f.fn_name;
        f.fn_symbol <- Some sym;
        set_property sym IsNonLvalue;
        let initial_ctx = prepare_ctx f.fn_loc (c, f.fn_parms) in
        ignore (bind_stmt (initial_ctx, f.fn_body));
        c
    | GlobalDecl g ->
        let (c, t) = bind_type (c, g.global_type) in
        let (c, sym) = 
          add_symbol c g.global_loc VariableKind t g.global_name in
        sym.sym_name <- c.ctx_module ^ "::" ^ g.global_name;
        set_property sym IsGlobalVar;
        g.global_symbol <- Some sym;
        c
    | ExternC x ->
        let (c, t) = bind_type (c, x.xc_type) in
        let (c, sym) = add_symbol c x.xc_loc VariableKind t x.xc_name in
        x.xc_symbol <- Some sym;
        c
    | TopLevelOpen m ->
        if m = c.ctx_module && c.ctx_iface then
          raise (Error "trying to open current module in interface")
        else (
          Gontmodule.open_module m;
          {c with ctx_opened_modules = m :: c.ctx_opened_modules}
        )
    | TypeDef ({td_name = n} as t) ->
        if t.td_module = "" then
          t.td_module <- c.ctx_module;
        (try 
          let s = find_type c "" n in
          if s.td_module = c.ctx_module then
            if s.td_type = AbstractType then 
              (* Allow first `type t' and then 'struct t {...}' *)
              ()
            else
              err (t.td_loc, xf "redefinition of type `%s'" n)
          else
            warn (t.td_loc, xf "type `%s' shadows `%s::%s'" 
                               n s.td_module s.td_name)
         with Not_found -> ());
        (* Allow recursive types. *)
        let c = {c with ctx_types = t :: c.ctx_types} in
        let c = add_fields (c, t) in
        let (c, tt) = bind_type (c, t.td_type) in
        t.td_type <- tt;
        c
    | FuncProto fp ->
        let (c, t) = bind_type (c, fp.fp_type) in
        let (c, sym) = 
          add_symbol c fp.fp_loc VariableKind t fp.fp_name in
        sym.sym_name <- c.ctx_module ^ "::" ^ fp.fp_name;
        if fp.fp_extern_name <> "" then
          set_property sym (ExternName fp.fp_extern_name);
        set_property sym IsNonLocal;
        fp.fp_symbol <- Some sym;
        c
    | ExceptionDecl ed ->
        let f t =
          match t with
          | TyVar _ -> err (ed.ed_loc, "sorry, exception types cannot be \
                                        polimorphic")
          | _ -> t in
        let (c, t) = bind_type (c, ed.ed_type) in
        let _ = type_walk f t in
        let (c, sym) = add_symbol c ed.ed_loc VariableKind t ed.ed_name in
        sym.sym_name <- c.ctx_module ^ "::" ^ ed.ed_name;
        ed.ed_sym <- Some sym;
        set_property sym IsException;
        c
    | DummyDecl -> c
    | SectionDecl sd ->
      begin
        match sd.sd_name with
        | "init" -> {c with ctx_init_body = sd.sd_body :: c.ctx_init_body}
        | "fini" -> {c with ctx_fini_body = sd.sd_body :: c.ctx_fini_body}
        | _ -> err (sd.sd_loc, xf "unknown section type `%s'" sd.sd_name)
      end
;;

let finalize_module out_ch c =
  Gontmodule.save_module out_ch c.ctx_module c.ctx_syms c.ctx_types

let init_stmts c = List.rev c.ctx_init_body
let fini_stmts c = List.rev c.ctx_fini_body
;;

(* Apply given function, of (exp -> 'a list) type, for all
 * expressions in a tree, returning flat list of results. *)
let rec for_all_exp exp_fnc s =
  let stmt_fnc s = for_all_exp exp_fnc s in
  let rec patt_divisor pl =
    match pl with
      x::xs -> 
        let {cc_stmt = st} = x in
        (stmt_fnc st) @ patt_divisor xs
    | [] -> [] in
  match s.raw_stmt with
  | Skip
  | VarDef _
  | Open _
  | Break _
  | ReturnVoid -> []
  | Exp e
  | BreakIfNot e
  | Raise e
  | Return e -> exp_fnc e
  | Block l -> (List.flatten (List.map stmt_fnc l))
  | Loop {ll_body = a; ll_inc = b}
  | Pair (a, b) -> (stmt_fnc a) @ (stmt_fnc b)
  | Switch (e, ptl) -> 
      (exp_fnc e) @ (patt_divisor ptl)
  | Try t -> 
      (stmt_fnc t.try_body) @ (patt_divisor t.try_with) @ 
      (stmt_fnc t.try_finally)
  | Let (pel, st) ->
      let rec pel_divisor pl =
        match pl with
          x::xs -> 
            let (_, e) = x in
            (exp_fnc e) @ pel_divisor xs
        | [] -> [] in
       (pel_divisor pel) @ (stmt_fnc st)        
  | If(e, s1, s2) -> exp_fnc e @ stmt_fnc s1 @ stmt_fnc s2
;;

(* Return list of symbols, at given scoping level, that needs to put into the
 * closure because they are referenced in stmts. *)
let closure lev s =
  let need_closure = ref false in
  let rec exp_closure lambda_lev e =
    let ec = exp_closure lambda_lev in
    match e.raw_exp with
        SimpleBinOp(op, e1, e2) -> (ec e1) @ (ec e2)
      | DotExp {de_expr = e1}
      | SimpleUnOp(_, e1) -> ec e1
      | Ref(r) ->
          (match r.ref_symbol with
              None -> assert false
            | Some sym ->
                if r.ref_level > 0 then
                  (if lambda_lev > 0 then need_closure := true;
                   if sym.sym_fnc_level = lev && 
                      not (has_property sym IsInClosure) then
                     (if has_property sym IsInPattern then
                        err(e.exp_loc, "cannot use pattern variable in local \
                            function");
                      set_property sym IsInClosure; [sym])
                   else
                     [])
                else
                  [])
      | Call(e1, es) ->
          (ec e1) @ (List.flatten (List.map ec es))
      | CrossModuleRef _
      | IntConst _
      | BoolConst _
      | FloatConst _
      | StringConst _
      | NullConst
      | VoidConst -> []
      | TupleExp (tup) ->
          let rec tup_apply tu =
            match tu with
              x::xs -> (ec x) @ (tup_apply xs)
            | [] -> [] in
          tup_apply tup
      | UnionExp (un) -> ec un.ur_exp
      | StructExp sis ->
         List.flatten (List.map (fun {si_expr = e} -> ec e) sis)
      | Lambda {lambda_body=s} ->
         for_all_exp (exp_closure (lambda_lev + 1)) s in
  (!need_closure, for_all_exp (exp_closure 0) s)
;;

