/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: parser.mly,v 1.17 2002/05/27 15:08:34 malekith Exp $
 *
 * Module implementation: Parser
 * Description: Parser in ocamlyacc for the Gont language.
 */
%{
open Gonttree
open Gontutil
;;

let locus () = 
  let fn = Lexer.get_file_name () in
  let (l, c) = Lexer.get_line_column fn (Parsing.symbol_start()) in
        {loc_file=fn;
         loc_char=c;
         loc_line=l}

let ice () = raise Parse_error

let make_exp raw =
  {raw_exp = raw;
   exp_type = None;
   exp_loc = locus ()}

let bin op e1 e2 =
  make_exp (SimpleBinOp(op, e1, e2))
  
let make_stmt raw =
  {raw_stmt = raw;
   stmt_loc = locus ()}
   
let make_ref id =
  make_exp (Ref {ref_name = id; 
                 ref_level = -1;
                 ref_symbol = None})
let make_pair a b =
  make_stmt (Pair (a, b))
  
let make_named_type (m, n) p =
  NamedType {nt_module = m;
             nt_name   = n;
             nt_parms  = p;
             nt_tydef  = None;
             nt_loc    = locus()}

let make_tyvars (r, a) = 
  let t = {ft_ret_type = r; 
           ft_parms = List.map (fun {parm_type=t} -> t) a; 
           ft_tyvars = []} in
  simple_fv (Func t)

let bare_fn_type (r, a) =
  let t = {ft_ret_type = r; ft_parms = a; ft_tyvars = []} in
  let fv = simple_fv (Func t) in
  Func {t with ft_tyvars = fv}
;;

let unary op e =
  {raw_exp = SimpleUnOp(op, e);
   exp_type = None;
   exp_loc = locus ()}
;;

%}

%token <string> INT
%token <string> FLOAT
%token <string> ID
%token <string> STRING
%token <string> TYVAR           /* 'a -> TYVAR("a") */
%token <char> CHAR
%token KW_INT KW_RETURN KW_VOID KW_SKIP KW_BOOL KW_FLOAT KW_IN
%token KW_EXTERN_C KW_STRING KW_FUN KW_IF KW_ELSE KW_WHILE KW_DO KW_FOR
%token KW_CASE KW_SWITCH KW_LET KW_EXPORT_AS KW_STRUCT KW_OPT_STRUCT
%token KW_UNION KW_TYPE KW_LOCAL KW_TYPEDEF KW_TRUE KW_FALSE KW_OPEN
%token KW_NULL KW_RAISE KW_TRY KW_EXCEPTION KW_BREAK KW_CONTINUE
%token KW_FINALLY KW_WITH KW_SECTION
%token COLON DOUBLECOLON ARROW LSQUARE RSQUARE UNDER DOT
%token LT GT GE LE EQ NE LTLT GTGT
%token PLUS MINUS TIMES DIV EQUAL MOD AT
%token LPAREN RPAREN LBRACE RBRACE SEMICOLON COMMA
%token PLUS_PLUS MINUS_MINUS

/* lowest precedence */
%right EQUAL
%left LT GT GE LE
%left EQ NE
%left PLUS MINUS AT
%left TIMES DIV MOD
%nonassoc UMINUS
%right PLUS_PLUS MINUS_MINUS
%nonassoc APPLICATION
%left DOT
/* highest precedence */

%start decl iface_decl            /* the entry points */

%type <Gonttree.decl> decl
%type <Gonttree.decl> iface_decl

%%

decl:
          func          { $1 }
        | global_var    { $1 }
        | extern_c      { $1 }
        | type_decl     { $1 }
        | exn_decl      { $1 }
        | KW_LOCAL func_proto    { $2 }
        | top_open      { $1 }
        | SEMICOLON     { DummyDecl }
        | section       { $1 }
;

iface_decl:
          func_proto    { $1 }
        | xc_func_proto { $1 }
        | global_var    { $1 }
        | exn_decl      { $1 }
        | type_decl     { $1 }
        | top_open      { $1 }
        | SEMICOLON     { DummyDecl }
;

id: 
        ID              { $1 }
;

qid:
          id                    { ("", $1) }
        | id DOUBLECOLON id     { ($1, $3) }
;

maybe_export_as:
         /* empty */            { "" }
       | KW_EXPORT_AS STRING    { $2 }
;

top_open:
        KW_OPEN id     { TopLevelOpen $2 }
;

exn_decl:
        KW_EXCEPTION type_exp id        { ExceptionDecl {ed_type = $2;
                                                         ed_name = $3;
                                                         ed_loc = locus();
                                                         ed_sym = None} }
;

type_var_list_content:
          TYVAR                   { [$1] }
        | TYVAR COMMA type_var_list_content { $1 :: $3 }
;
        
type_var_list:
          /* empty */           { [] }
        | LT GT                 { [] }
        | LT type_var_list_content GT   { $2 }
;

maybe_semicolons:
          /* nothing */                 { [] }
        | maybe_semicolons SEMICOLON    { [] }
;

record_type_fld:
        type_exp qid maybe_semicolons    { let (m, n) = $2 in
                                           {fr_module = m;
                                            fr_name = n;
                                            fr_type = $1;
                                            fr_sym = None;
                                            fr_loc = locus()} }
;

record_type_flds:
          record_type_fld                       { [$1] }
        | record_type_fld record_type_flds      { $1::$2 }
;

record_type_content:
          type_var_list id LBRACE record_type_flds RBRACE 
          { ($4, {td_module = "";
                  td_name = $2;
                  td_tyvars = $1;
                  td_type = Void;
                  td_loc = locus()}) }
;

type_decl:
          KW_STRUCT record_type_content {
            let (l, r) = $2 in 
            TypeDef {r with td_type = Struct l} }
        | KW_OPT_STRUCT record_type_content {
            let (l, r) = $2 in 
            TypeDef {r with td_type = OptStruct l} }
        | KW_UNION record_type_content {
            let (l, r) = $2 in 
            TypeDef {r with td_type = Union l} }
        | KW_TYPEDEF type_exp type_var_list id {
             TypeDef {td_module = "";
                      td_name = $4;
                      td_tyvars = $3;
                      td_type = $2;
                      td_loc = locus()} }
        | KW_TYPE type_var_list id {
             TypeDef {td_module = "";
                      td_name = $3;
                      td_tyvars = $2;
                      td_type = AbstractType;
                      td_loc = locus()} }
;

func:
       type_exp id LPAREN func_args RPAREN maybe_export_as lambda_body
                { FuncDecl {fn_loc = locus ();
                            fn_ret_type = $1; 
                            fn_name = $2;
                            fn_parms = $4;
                            fn_body = $7;
                            fn_symbol = None;
                            fn_export_as = $6;
                            fn_tyvars = make_tyvars ($1, $4)} }
;

section:
        KW_SECTION id lambda_body
                { SectionDecl {sd_name = $2;
                               sd_body = $3;
                               sd_loc = locus()} }
;

global_var:
       type_exp id SEMICOLON
                { GlobalDecl {global_loc = locus ();
                              global_type = $1;
                              global_name = $2;
                              global_symbol = None} }
;

/* from now on ne_ == non_empty_ */

func_arg:
        type_exp id             { {parm_type = $1; 
                                   parm_name = $2; 
                                   parm_symbol = None} }
;

ne_func_args:
          func_arg                      { [$1] }
        | func_arg COMMA ne_func_args   { $1 :: $3 }
;

func_args:
          /* empty */    { [] }
        | ne_func_args   { $1 }
;

proto_arg:
          type_exp              { $1 }
        | type_exp id           { $1 }
;

ne_proto_args:
          proto_arg                             { [$1] }
        | proto_arg COMMA ne_proto_args         { $1 :: $3 }
;

proto_args:
          /* empty */           { [] }
        | ne_proto_args         { $1 }
;

proto:
        type_exp id LPAREN proto_args RPAREN
                { ($1, $2, $4) }
;

ne_expr_list:
          expr                           { [$1] }
        | expr COMMA ne_expr_list        { $1 :: $3 }
;

expr_list:
          /* empty */    { [] }
        | ne_expr_list   { $1 }
;

extern_c:
          KW_EXTERN_C proto SEMICOLON {
            let (t, n, a) = $2 in
              ExternC {xc_type = bare_fn_type (t, a);
                       xc_name = n;
                       xc_loc = locus ();
                       xc_symbol = None} }
        | KW_EXTERN_C type_exp id SEMICOLON {
              ExternC {xc_type = $2;
                       xc_name = $3;
                       xc_loc = locus ();
                       xc_symbol = None} }
;

func_proto:
        proto SEMICOLON {
            let (t, n, a) = $1 in
              FuncProto {fp_type = bare_fn_type (t, a);
                         fp_name = n;
                         fp_loc = locus ();
                         fp_symbol = None;
                         fp_extern_name = ""} }
;

xc_func_proto:
        proto EQUAL STRING SEMICOLON {
            let (t, n, a) = $1 in
              FuncProto {fp_type = bare_fn_type (t, a);
                         fp_name = n;
                         fp_loc = locus ();
                         fp_symbol = None;
                         fp_extern_name = $3} }
;

block:
        LBRACE stmts RBRACE { 
          make_stmt (match $2 with
                        [] -> Skip
                      | _ -> Block $2) 
        }
;

stmts:
          /* empty */      { [] }
        | stmt stmts       { $1::$2 }
;

non_if_stmt:
          normal_stmt SEMICOLON         { $1 }
        | block                         { $1 }
        | func_def                      { $1 }
        | loop_stmt                     { $1 }
        | pattern_matching_stmt         { $1 }
        | try_stmt                      { $1 }
;

if_stmt:
          KW_IF LPAREN expr RPAREN if_stmt {
              make_stmt (If ($3, $5, make_stmt Skip)) 
        }
        | KW_IF LPAREN expr RPAREN non_if_stmt 
                { make_stmt (If ($3, $5, make_stmt Skip)) }
;

if_else_stmt:
        KW_IF LPAREN expr RPAREN non_if_stmt KW_ELSE non_if_stmt {
            make_stmt (If ($3, $5, $7))
        }
; 

stmt:
          non_if_stmt                   { $1 }
        | if_stmt                       { $1 }
        | if_else_stmt                  { $1 } 
;

id_expr:
          id                    { ($1, None) }
        | id EQUAL expr         { ($1, Some $3) }
;

id_expr_list:
          id_expr                       { [$1] }
        | id_expr COMMA id_expr_list    { $1 :: $3 }
;

var_def:
          type_exp id_expr_list {
             let rec mk_defs lst =
               match lst with
               | (id, x)::xs ->
                 let def =
                    let d = 
                      make_stmt (VarDef {var_type = $1;
                                         var_name = id;
                                         var_symbol = None}) in
                    match x with
                    | None -> d
                    | Some e -> 
                       let assign = SimpleBinOp(Assign, (make_ref id), e) in
                       (* Need to set location of Assign to location of e,
                        * otherwise we would get really confusing error
                        * message in case of multiline variable definition. *)
                       let ini = make_stmt (Exp ({raw_exp = assign;
                                                  exp_type = None;
                                                  exp_loc = e.exp_loc})) in
                       make_pair d ini in
                  begin
                    match xs with
                    | [] -> def
                    | xs -> make_pair def (mk_defs xs)
                  end
               | [] -> assert false in
             mk_defs $2
        }
;

func_def:
        type_exp id LPAREN func_args RPAREN lambda_body {
           let ty = make_fn_type ($1, $4) in
           let tyvars = match ty with Func {ft_tyvars = x} -> x 
                                    | _ -> assert false in
           let body = make_exp (Lambda {lambda_ret_type = $1;
                                        lambda_parms = $4;
                                        lambda_body = $6;
                                        lambda_tyvars = tyvars
                                        }) in
           let def = make_stmt (VarDef {var_type = ty; 
                                        var_name = $2; 
                                        var_symbol = None}) in
           let init = bin Assign (make_ref $2) body in
           make_pair def (make_stmt (Exp init))
        }
;

for_exp1:
          stmt_expr             { make_stmt (Exp $1) }
        | var_def               { $1 }
        | KW_SKIP               { make_stmt Skip }
        | /* empty */           { make_stmt Skip }
;

for_exp2:
          expr                  { $1 }
        | /* empty */           { make_exp (BoolConst true) }
;

for_exp3:
          stmt_expr             { make_stmt (Exp $1) }
        | KW_SKIP               { make_stmt Skip }
        | /* empty */           { make_stmt Skip }
;        

maybe_label:
        /* empty */             { "" }
      | id COLON                { $1 }
;

maybe_id:
        /* empty */             { "" }
      | id                      { $1 }
;

loop_stmt:
         maybe_label KW_WHILE LPAREN expr RPAREN stmt {
              let cond = make_stmt (BreakIfNot $4) in
              make_stmt(Loop {ll_body = make_pair cond $6;
                              ll_label = $1;
                              ll_sym = None;
                              ll_inc = make_stmt Skip})
        }
        | maybe_label KW_DO block KW_WHILE LPAREN expr RPAREN SEMICOLON {
              let cond = make_stmt (BreakIfNot $6) in
              make_stmt(Loop {ll_body = make_pair $3 cond;
                              ll_label = $1;
                              ll_sym = None;
                              ll_inc = make_stmt Skip})
        }
        | maybe_label KW_FOR LPAREN 
                for_exp1 SEMICOLON 
                for_exp2 SEMICOLON 
                for_exp3 RPAREN stmt {
              let cond = make_stmt (BreakIfNot $6) in
              let loop = Loop {ll_body = make_pair cond $10;
                               ll_inc = $8;
                               ll_sym = None;
                               ll_label = $1} in
              make_stmt (Block [$4; make_stmt loop])
        }
;

normal_stmt:
          stmt_expr             { make_stmt (Exp $1) }
        | KW_RETURN             { make_stmt ReturnVoid }
        | KW_RETURN expr        { make_stmt (Return $2) }
        | KW_SKIP               { make_stmt Skip }
        | var_def               { $1 }
        | KW_RAISE expr         { make_stmt (Raise $2) }
        | KW_BREAK maybe_id     { make_stmt (Break {br_label = $2;
                                                    br_sym = None;
                                                    br_continue = false}) }
        | KW_CONTINUE maybe_id  { make_stmt (Break {br_label = $2;
                                                    br_sym = None;
                                                    br_continue = true}) }
;

variable:
        qid { let (m, n) = $1 in
                if m = "" then 
  (*let fn = Lexer.get_file_name () in
  let (l, c) = Lexer.get_line_column fn (Parsing.symbol_start()) in
                  (Printf.printf "variable %s %s@%d:%d\n" 
                    n fn l c    ; *) 
                  make_ref n
                else
                  make_exp (CrossModuleRef {xmr_module = m; 
                                            xmr_name   = n;
                                            xmr_symbol = None})
            }
;

tuple_content:
          expr COMMA expr               { [$1; $3] }
        | expr COMMA tuple_content      { $1 :: $3 }
;

tuple:
        LSQUARE tuple_content RSQUARE     { make_exp (TupleExp $2) }
;

/* expression that might be use to form expression stmt */
stmt_expr:
          call %prec APPLICATION  { $1 }
        | expr EQUAL expr         { bin Assign $1 $3 }
        | MINUS_MINUS expr        { unary PreDec $2 }
        | PLUS_PLUS expr          { unary PreInc $2 }
        | expr MINUS_MINUS        { unary PostDec $1 }
        | expr PLUS_PLUS          { unary PostInc $1 }
;

pattern_expr:
          primary                 { $1 }
        | pattern_expr PLUS pattern_expr          { bin Plus $1 $3 }
        | pattern_expr MINUS pattern_expr         { bin Minus $1 $3 }
        | pattern_expr TIMES pattern_expr         { bin Mult $1 $3 }
        | pattern_expr DIV pattern_expr           { bin Div $1 $3 }
        | pattern_expr MOD pattern_expr           { bin Mod $1 $3 }
        | pattern_expr LT pattern_expr            { bin Lt $1 $3 }
        | pattern_expr GT pattern_expr            { bin Gt $1 $3 }
        | pattern_expr GE pattern_expr            { bin Ge $1 $3 }
        | pattern_expr LE pattern_expr            { bin Le $1 $3 }
        | pattern_expr EQ pattern_expr            { bin Eq $1 $3 }
        | pattern_expr NE pattern_expr            { bin Ne $1 $3 }
        | MINUS pattern_expr %prec UMINUS { unary Negate $2 }
;

union_expr:
        /* union constructor */
          qid LSQUARE RSQUARE { 
              let (m, n) = $1 in
              make_exp (UnionExp {
                        ur_module = m;
                        ur_name = n;
                        ur_exp = make_exp VoidConst;
                        ur_sym = None
                      }) 
         }
        | qid LSQUARE expr RSQUARE { 
              let (m, n) = $1 in
              make_exp (UnionExp {
                        ur_module = m;
                        ur_name = n;
                        ur_exp = $3;
                        ur_sym = None
                      }) 
         }
        | qid tuple { 
              let (m, n) = $1 in
              make_exp (UnionExp {
                        ur_module = m;
                        ur_name = n;
                        ur_exp = $2;
                        ur_sym = None
                      }) 
         }
;

expr:
          primary                 { $1 }
        | expr PLUS expr          { bin Plus $1 $3 }
        | expr MINUS expr         { bin Minus $1 $3 }
        | expr TIMES expr         { bin Mult $1 $3 }
        | expr DIV expr           { bin Div $1 $3 }
        | expr MOD expr           { bin Mod $1 $3 }
        | expr LT expr            { bin Lt $1 $3 }
        | expr GT expr            { bin Gt $1 $3 }
        | expr GE expr            { bin Ge $1 $3 }
        | expr LE expr            { bin Le $1 $3 }
        | expr EQ expr            { bin Eq $1 $3 }
        | expr NE expr            { bin Ne $1 $3 }
        | MINUS expr %prec UMINUS { unary Negate $2 }
        | tuple                   { $1 }
        | stmt_expr               { $1 }
        | struct_expr             { $1 }
        | union_expr              { $1 }
;

primary:
          INT                     { make_exp (IntConst $1) }
        | FLOAT                   { make_exp (FloatConst $1) }
        | STRING                  { make_exp (StringConst $1) }
        | KW_TRUE                 { make_exp (BoolConst true) }
        | KW_FALSE                { make_exp (BoolConst false) }
        | variable                { $1 }
        | LPAREN expr RPAREN      { $2 }
        | lambda                  { $1 }
        | KW_NULL                 { make_exp NullConst }
        | primary DOT id          { make_exp (DotExp {de_expr = $1;
                                                      de_id = $3;
                                                      de_sym = None}) }
;

struct_init:
        qid EQUAL expr            { let (m, n) = $1 in {si_module = m;
                                                        si_name = n;
                                                        si_expr = $3;
                                                        si_loc = locus();
                                                        si_sym = None} }
;

struct_expr_content:
          struct_init                             { [$1] }
        | struct_init COMMA struct_expr_content   { $1 :: $3 }
;

struct_expr:
        LPAREN LBRACE struct_expr_content RBRACE RPAREN
                { make_exp (StructExp $3) }
;

call:
        primary LPAREN expr_list RPAREN  { make_exp (Call($1, $3)) }
;
        
simple_type:
          KW_VOID       { Void }
        | KW_INT        { Int }
        | KW_BOOL       { Bool }
        | KW_FLOAT      { Float }
        | KW_STRING     { String }
;

lambda_header:
           KW_FUN type_exp LPAREN func_args RPAREN
                { ($4, $2) }
;

lambda_body:
          block { $1 }
        | LPAREN expr RPAREN {
                let e = make_stmt (Return $2) in
                make_stmt (Block([e]))
        }
;

lambda:
          lambda_header lambda_body {
                let (p, r) = $1 in
                let tyvars = make_tyvars (r, p) in
                make_exp (Lambda {lambda_ret_type = r;
                                  lambda_parms = p;
                                  lambda_body = $2;
                                  lambda_tyvars = tyvars}) }
;

function_type:
           TIMES type_exp LPAREN proto_args RPAREN 
                { bare_fn_type ($2, $4) }
         | TIMES LPAREN type_exp LPAREN proto_args RPAREN RPAREN
                { bare_fn_type ($3, $5) }
;

tuple_ty:
          type_exp COMMA ne_ty_list     { $1 :: $3 }
;

ne_ty_list:
          type_exp                      { [$1] }
        | type_exp COMMA ne_ty_list     { $1 :: $3 }
;

ty_list:
          /* empty */                   { [] }
        | ne_ty_list                    { $1 }
;

xtype_exp: 
          simple_type                   { $1 }
        | function_type                 { $1 }
        | TYVAR                         { TyVar($1) }
        | LT ty_list GT qid             { make_named_type $4 $2 }
        /* this is hack to allow <<t1>x>y */
        | LTLT ne_ty_list GT qid GT qid 
          { make_named_type $6 [make_named_type $4 $2] }
        | LTLT ne_ty_list GT qid COMMA ne_ty_list GT qid 
          { make_named_type $8 ((make_named_type $4 $2) :: $6) }
        | TIMES LSQUARE tuple_ty RSQUARE        { Tuple $3 }
;

type_exp:
          LPAREN xtype_exp RPAREN       { $2 }
        | qid                           { make_named_type $1 [] }
        | xtype_exp                     { $1 }
;


ne_pattern_list:
          pattern COMMA ne_pattern_list   { $1 :: $3 }
        | pattern COMMA pattern           { [$1; $3] }
;

pattern_list:
          /* empty */                     { [] }
        | ne_pattern_list                 { $1 }
        | pattern                         { [$1] }
;

pattern:
        | UNDER                          { VoidPattern }
        | pattern_expr                   { ExprPattern $1 }
        | qid LSQUARE RSQUARE { 
          let (m, n) = $1 in
            UnionPattern {
              upr_module = m;
              upr_name = n;
              upr_pattern = {raw_pattern = 
                             TuplePattern {tpr_patterns = []};
                             pat_loc = locus ()};
              upr_sym = None
            } 
        }
        | qid LSQUARE pattern RSQUARE { 
          let (m, n) = $1 in
            UnionPattern {
              upr_module = m;
              upr_name = n;
              upr_pattern = {raw_pattern = $3; pat_loc = locus()};
              upr_sym = None
            } 
        }
        | qid LSQUARE ne_pattern_list RSQUARE { 
          let (m, n) = $1 in
            UnionPattern {
              upr_module = m;
              upr_name = n;
              upr_pattern = {raw_pattern = TuplePattern {
                    tpr_patterns = List.map (fun x -> {raw_pattern = x; 
                                                       pat_loc = locus()}) $3
                  }; pat_loc = locus()};
              upr_sym = None
            } 
        }
        | LSQUARE pattern_list RSQUARE 
            { TuplePattern {
                tpr_patterns = List.map (fun x -> {raw_pattern = x; 
                                                   pat_loc = locus()}) $2
              } 
            }
;

case_clause:
        KW_CASE pattern COLON stmts     { {cc_pattern = {raw_pattern = $2; 
                                                         pat_loc = locus ()};
                                           cc_stmt = make_stmt (Block $4)} }
;

case_list:
          case_clause case_list           { $1 :: $2 }
        | /* empty */                     { [] }
;

pattern_eq:
          pattern EQUAL expr            { ({raw_pattern = $1; 
                                            pat_loc = locus()}, $3) }
;

pattern_eq_list:
          pattern_eq                            { [$1] }
        | pattern_eq COMMA pattern_eq_list      { $1 :: $3 }
;

maybe_finally:
          KW_FINALLY stmt                 { $2 }
        | /* nothing */                   { make_stmt Skip }
;

maybe_with:
          KW_WITH LBRACE case_list RBRACE       { $3 }
        | /* nothing */                         { [] }
;

try_stmt:
        KW_TRY block maybe_with maybe_finally    
                        { make_stmt (Try {try_body = $2;
                                          try_with = $3;
                                          try_finally = $4}) }
;

pattern_matching_stmt:
          KW_LET pattern_eq_list KW_IN block    { make_stmt 
                                                    (Let ( $2, $4))}
        | KW_SWITCH expr LBRACE case_list RBRACE  { make_stmt 
                                                      (Switch ($2, $4)) }
;
